import os
import random
from pathlib import Path

from dotenv import load_dotenv

from countergen.tools.api_utils import ApiConfig

load_dotenv()

verbose = int(os.environ.get("CDG_VERBOSE", "2"))

MODULE_PATH = str(Path(__file__).parent)  # To load internal data

DEFAULT_API_BASE_URL = "https://api.openai.com/v1"

apiconfig = ApiConfig(
    key=os.environ.get("OPENAI_API_KEY", None),
    base_url=os.environ.get("OPENAI_API_BASE_URL", DEFAULT_API_BASE_URL),
)


def seed(s: int):
    random.seed(s)



# From Fryer, 2022 https://aclanthology.org/2022.woah-1.20.pdf
# Adapted to be usable with InstructGPT
#%%
from typing import Dict, Optional, Tuple
import countergen
from countergen.tools.api_utils import ApiConfig

import openai
from attrs import define
from countergen.tools.utils import estimate_paraphrase_length
from countergen.types import Augmenter, Category, Input
import countergen.config

DEFAULT_AUGMENTERS = {
    "gender": {
        "male": "Rewrite it to be about a man/about men.",
        "female": "Rewrite it to be about a woman/about women.",
    },
}

DEFAULT_PROMPT = """0: Here is some text: {When the doctor asked Linda to take the medicine, he smiled and gave her a lollipop.}. Rewrite it to be more scary.
1: {When the doctor told Linda to take the medicine, there had been a malicious gleam in her eye that Linda didn’t like at all.}
0: Here is some text: {they asked loudly, over the sound of the train.}. Rewrite it to be more intense.
1: {they yelled aggressively, over the clanging of the train.}
0: Here is some text: {When Mohammed left the theatre, it was already dark out}. Rewrite it to be more about the movie itself.
1: {The movie was longer than Mohammed had expected, and despite the excellent ratings he was a bit disappointed when he left the theatre.}
0: Here is some text: {next to the path}. Rewrite it to be about France.
1: {next to la Siene}
0: Here is some text: {The man stood outside the grocery store, ringing the bell.}. Rewrite it to be about clowns.
1: {The man stood outside the circus, holding a bunch of balloons.}
0: Here is some text: {the bell ringing}. Rewrite it to be more flowery.
1: {the peales of the jangling bell}
0: Here is some text: {against the tree}. Rewrite it to include the word “snow”.
1: {against the snow-covered bark of the tree}’
0: Here is some text: {__input__}. __instruction__
1: {"""


@define
class LlmdAugmenter(Augmenter):
    """Augmenter that does word substituion between its two categories.

    "words" are defined by the word_regex expression.

    From Fryer 2022, https://aclanthology.org/2022.woah-1.20.pdf"""

    categories_instructions: Dict[Category, str]
    prompt_template: str = DEFAULT_PROMPT
    engine: str = "text-davinci-003"
    apiconfig: Optional[ApiConfig] = None

    @classmethod
    def from_default(cls, name: str) -> "LlmdAugmenter":
        """Load one of the defaults datasets from "DEFAULT_AUGMENTERS"."""
        if name not in DEFAULT_AUGMENTERS:
            raise ValueError(f"{name} not a valid default augmenter. Choose one in {set(DEFAULT_AUGMENTERS.keys())}")
        return LlmdAugmenter(DEFAULT_AUGMENTERS[name])

    @property
    def categories(self) -> Tuple[Category, ...]:
        return tuple(self.categories_instructions.keys())

    def transform(self, inp: Input, to: Category) -> Input:
        apiconfig = self.apiconfig or countergen.config.apiconfig

        instruction = self.categories_instructions[to]
        prompt = self.prompt_template.replace("__input__", inp).replace("__instruction__", instruction)

        completion = openai.Completion.create(
            engine=self.engine,
            prompt=prompt,
            max_tokens=estimate_paraphrase_length(inp),
            temperature=1,
            top_p=0.7,  # LLM-D has top_k=40, but not available
            stream=False,
            **apiconfig.get_config(),
        )["choices"][0]["text"]

        return completion.split("}")[0]



from functools import lru_cache
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Sequence, Tuple, TypeVar, Union

import cattrs
import countergen.config
import openai
from tqdm import tqdm  # type: ignore

T = TypeVar("T")


def other(t: Tuple[T, T], x: T) -> T:
    if x == t[0]:
        if x == t[1]:
            raise ValueError(f"{t} contains two copies of {x}")
        return t[1]
    if x != t[1]:
        raise ValueError(f"{t} does not contain {x}")
    return t[0]


def all_same(l: Sequence[Any]) -> bool:
    if not l:
        return True
    return all(x == l[0] for x in l[1:])


def maybe_tqdm(it: Iterable[T], do_tqdm: bool = False, **kwargs) -> Iterable[T]:
    if do_tqdm:
        return tqdm(it, **kwargs)
    else:
        return it


def estimate_paraphrase_length(text: str):
    average_token_length = 3
    safety_margin = 50
    return len(text) // average_token_length + safety_margin


class FromAndToJson:
    """Add a from_json_dict & to_json_dict functions.

    The class is expected to be created using @attrs.define."""

    @classmethod
    def from_json_dict(cls, json_dict: Mapping[str, Any]):
        return cattrs.structure(json_dict, cls)

    def to_json_dict(self) -> Mapping[str, Any]:
        return cattrs.unstructure(self, self.__class__)


def unwrap_float(x: Union[float, Any], err_msg="Value should be a float!") -> float:
    if not isinstance(x, float):
        raise ValueError(err_msg)
    return x


def unwrap_list_of_floats(x: Union[List[float], Any], err_msg="Value should be a list of floats!") -> List[float]:
    if not isinstance(x, list):
        raise ValueError(err_msg)
    if not all(isinstance(y, float) for y in x):
        raise ValueError(err_msg)
    return x



from countergen.tools.cli import run


if __name__ == "__main__":
    run()

    # python -m countergen augment countergen\data\datasets\tiny-test.jsonl countergen\data\augdatasets\tiny-test.jsonl gender
    # python -m countergen augment countergen\data\datasets\twitter-sentiment.jsonl countergen\data\augdatasets\twitter-sentiment.jsonl gender
    # python -m countergen augment countergen\data\datasets\doublebind.jsonl countergen\data\augdatasets\doublebind.jsonl gender
    # python -m countergen augment countergen\data\datasets\hate-test.jsonl countergen\data\augdatasets\hate-test.jsonl gender
    # python -m countergen evaluate countergen\data\augdatasets\doublebind.jsonl
    # python -m countergen evaluate countergen\data\augdatasets\doublebind.jsonl --model-name davinci
    # python -m countergen augment countergen\data\datasets\doublebind-positive.jsonl countergen\data\augdatasets\doublebind-positive.jsonl paraphrase gender
    # python -m countergen augment countergen\data\datasets\male-stereotypes.jsonl countergen\data\augdatasets\male-stereotypes.jsonl gender
    # python -m countergen augment countergen\data\datasets\female-stereotypes.jsonl countergen\data\augdatasets\female-stereotypes.jsonl gender



from collections import defaultdict
from typing import Any, DefaultDict, Iterable, List, Mapping, Optional, Sequence, TextIO, Tuple, TypeVar

import numpy as np
from attrs import define

from countergen.tools.plot_utils import plot_mutli_bars
from countergen.types import AugmentedSample, Category, Input, Performance, Results, Aggregator, Outputs
from countergen.tools.math_utils import geometric_mean, mean
from countergen.tools.utils import FromAndToJson, unwrap_float, unwrap_list_of_floats


@define
class AveragePerformancePerCategory(Aggregator):
    """Compute the average performance for each category."""

    use_geometric_mean: bool = False

    def __call__(self, performances: Results) -> Mapping[Category, float]:
        performances_by_category: DefaultDict[Category, List[float]] = defaultdict(lambda: [])
        for sample_perfs in performances:
            for perf, categories in sample_perfs:
                for c in categories:
                    if isinstance(perf, float):
                        performances_by_category[c].append(perf)
                    else:  # perf is list of floats
                        for p in perf:
                            performances_by_category[c].append(p)

        mean_ = geometric_mean if self.use_geometric_mean else mean
        avg_performances_by_category = {c: mean_(perfs) for c, perfs in performances_by_category.items()}
        return avg_performances_by_category

    def save_aggregation(self, aggregate: Mapping[Category, float], file: Optional[TextIO] = None):
        print("Average performance per category:", file=file)
        for c, perf in aggregate.items():
            print(f"{c}: {perf:.3e}", file=file)

    def load_aggregation(self, file: TextIO) -> Mapping[Category, float]:
        lines = file.readlines()
        r = {}
        for l in lines[1:]:
            c, p = l.split(": ")
            r[c] = float(p)
        return r

    def display(self, aggregates: Mapping[str, Mapping[Category, float]]):
        plot_mutli_bars(aggregates, xlabel="Model name", ylabel="Performance", title="Performance by model & category")


@define
class Stats(FromAndToJson):
    mean: float
    uncertainty_2sig: float

    def __str__(self) -> str:
        return f"{self.mean:.3e} +- {self.uncertainty_2sig:.3e}"

    @classmethod
    def from_str(cls, s: str) -> "Stats":
        m, u = s.split(" +- ")
        return Stats(float(m), float(u))

    @classmethod
    def from_seq(cls, s: Sequence[float]) -> "Stats":
        assert s, "empty seq"
        a = np.array(s)
        uncertainty_2sig = 2 * a.std() / np.sqrt(len(s))
        return Stats(mean=a.mean(), uncertainty_2sig=uncertainty_2sig)


@define
class PerformanceStatsPerCategory(Aggregator):
    """Compute performance mean and the 2 sigma uncertainty over mean for each category."""

    def __call__(self, performances: Results) -> Mapping[Category, Stats]:
        performances_by_category: DefaultDict[Category, List[float]] = defaultdict(lambda: [])
        for sample_perfs in performances:
            for perf, categories in sample_perfs:
                for c in categories:
                    if isinstance(perf, float):
                        performances_by_category[c].append(perf)
                    else:  # perf is list of floats
                        for p in perf:
                            performances_by_category[c].append(p)

        aggregate = {c: Stats.from_seq(perfs) for c, perfs in performances_by_category.items()}
        return aggregate

    def save_aggregation(self, aggregate: Mapping[Category, Stats], file: Optional[TextIO] = None):
        print("Average performance per category:", file=file)
        for c, perf in aggregate.items():
            print(f"{c}: {perf}", file=file)

    def load_aggregation(self, file: TextIO) -> Mapping[Category, Stats]:
        lines = file.readlines()
        r = {}
        for l in lines[1:]:
            c, p = l.split(": ")
            r[c] = Stats.from_str(p)
        return r

    def display(self, aggregates: Mapping[str, Mapping[Category, Stats]]):
        means = {n: {c: s.mean for c, s in d.items()} for n, d in aggregates.items()}
        errs = {n: {c: s.uncertainty_2sig for c, s in d.items()} for n, d in aggregates.items()}
        plot_mutli_bars(
            means,
            xlabel="Model name",
            ylabel="Performance",
            title="Performance by model & category",
            err_by_type_by_cat=errs,
        )


@define
class DifferenceStats(Aggregator):
    """Compute performance mean and the 2 sigma uncertainty (relative) difference of the performance in each samples.

    Return a positive mean if category1 has higher performance that category2, and a negative one otherwise.

    If a sample has mutliple variations of the same category, compute the mean performance of the category.

    Excepts performance to be a float."""

    category1: Category
    category2: Category
    relative_difference: bool = True

    def __call__(self, performances: Results) -> Stats:
        differences = []
        for sample_perfs in performances:
            cat1_perfs = [unwrap_float(perf) for perf, categories in sample_perfs if self.category1 in categories]
            cat2_perfs = [unwrap_float(perf) for perf, categories in sample_perfs if self.category2 in categories]
            if cat1_perfs and cat2_perfs:
                diff = mean(cat1_perfs) - mean(cat2_perfs)
                if self.relative_difference:
                    diff /= max(mean(cat1_perfs), mean(cat2_perfs))
                differences.append(diff)
        return Stats.from_seq(differences)

    def save_aggregation(self, aggregate: Stats, file: Optional[TextIO] = None):
        relative = "relative " if self.relative_difference else ""
        print(
            f"The {relative}performance between category {self.category1} and category {self.category2}:",
            file=file,
        )
        print(f"{aggregate}", file=file)

    def load_aggregation(self, file: TextIO) -> Stats:
        lines = file.readlines()
        return Stats.from_str(lines[1])


OutlierData = Tuple[Input, Outputs, Tuple[Category, ...], Performance]


@define
class OutliersAggregator(Aggregator):
    """Return the variations with the biggest (relative) performance gap."""

    aug_samples: Iterable[AugmentedSample]  #: Contains data about the inputs used
    top_k: int = 5
    perf_per_output: bool = False  #: Performance is expected to be a list of float, one per output
    use_relative_performance: bool = True
    epsilon: float = 1e-10

    def __call__(self, performances: Results) -> List[Tuple[OutlierData, OutlierData]]:
        possibles_outliers: List[Tuple[OutlierData, OutlierData, float]] = []
        for aug_sample, variations_perf in zip(self.aug_samples, performances):
            if len(variations_perf) < 2:
                continue

            if not self.perf_per_output:
                outliers_data = list(
                    zip(
                        [v.text for v in aug_sample.get_variations()],
                        [aug_sample.outputs for _ in range(len(aug_sample.get_variations()))],
                        [cats for _, cats in variations_perf],
                        [unwrap_float(perf) for perf, _ in variations_perf],
                    )
                )

                sorted_outliers = sorted(outliers_data, key=lambda t: t[3])
                small_outlier = sorted_outliers[0]
                big_outlier = sorted_outliers[-1]
                possibles_outliers.append((small_outlier, big_outlier, self.gap(small_outlier[3], big_outlier[3])))
            else:
                just_perfs = [unwrap_list_of_floats(p) for p, _ in variations_perf]
                for i, output in enumerate(aug_sample.outputs):

                    outliers_data = list(
                        zip(
                            [v.text for v in aug_sample.get_variations()],
                            [
                                [
                                    output,
                                ]
                                for _ in range(len(aug_sample.get_variations()))
                            ],
                            [cats for _, cats in variations_perf],
                            [perf[i] for perf in just_perfs],
                        )
                    )

                    sorted_outliers = sorted(outliers_data, key=lambda t: t[3])
                    small_outlier = sorted_outliers[0]
                    big_outlier = sorted_outliers[-1]
                    possibles_outliers.append((small_outlier, big_outlier, self.gap(small_outlier[3], big_outlier[3])))

        best_outliers = sorted(possibles_outliers, key=lambda t: t[2], reverse=True)
        return [(small, big) for small, big, _ in best_outliers][: self.top_k]

    def display(self, aggregates: Mapping[str, List[Tuple[OutlierData, OutlierData]]]):
        for model_name, aggregate in aggregates.items():
            print(f"Biggest performance gaps for {model_name}:\n")
            for (inp1, outs1, cats1, perf1), (inp2, outs2, cats2, perf2) in aggregate:
                print(f"Performance={perf1:.3e} on input in categories {cats1}: {inp1} -> {outs1}")
                print(f"Performance={perf2:.3e} on input in categories {cats2}: {inp2} -> {outs2}\n")
            print("-----\n")

    def gap(self, small_perf: float, big_perf: float):
        delta = big_perf - small_perf
        if not self.use_relative_performance:
            return delta

        if small_perf < 0:
            raise ValueError("Can't use relative performance when performance is negative.")

        return delta / (big_perf + self.epsilon)


class BiasFromProbsAggregator(Aggregator):
    """Return average of log the biggest relative performance gap per sample accross all outputs and variations.

    Metric from BigBench's "Social Bias from Sentence Probability" experiment benchmark
    https://github.com/google/BIG-bench/tree/main/bigbench/benchmark_tasks/bias_from_probabilities"""

    categories: Optional[str] = None  #: Which categories to take into account
    epsilon: float = 1e-10
    perf_per_output: bool = False  #: Performance is expected to be a list of float, one per output

    def __call__(self, performances: Results) -> float:
        smallest_ratios = []
        for variations_perf in performances:
            if len(variations_perf) < 2:
                continue

            if self.perf_per_output:
                just_perfs = [unwrap_list_of_floats(p) for p, _ in variations_perf]
                smallest_ratios_in_sample = []
                for i in range(len(just_perfs[0])):
                    perf_by_output = [p[i] for p in just_perfs]
                    smallest_ratios_in_sample.append(min(perf_by_output) / max(perf_by_output))
                smallest_ratios.append(min(smallest_ratios_in_sample))
            else:
                just_perf = [unwrap_float(p) for p, _ in variations_perf]
                smallest_ratios.append(min(just_perf) / max(just_perf))

        return mean([np.log(g) for g in smallest_ratios])



from countergen.augmentation.data_augmentation import (
    SimpleAugmentedSample,
    AugmentedDataset,
    Dataset,
    DEFAULT_DS_PATHS,
    DEFAULT_AUGMENTED_DS_PATHS,
    generate_all_variations,
)
from countergen.augmentation.llmd_augmenter import LlmdAugmenter, DEFAULT_AUGMENTERS
from countergen.augmentation.paraphraser import LlmParaphraser
from countergen.augmentation.simple_augmenter import SimpleAugmenter, DEFAULT_CONVERTERS_PATHS
from countergen.evaluation import aggregators
from countergen.evaluation.evaluation import evaluate, evaluate_and_print, evaluate_and_save, compute_performances
from countergen.evaluation.generative_models import api_to_generative_model, get_generative_model_evaluator
from countergen.tools.cli import run
from countergen.types import (
    Input,
    Outputs,
    Sample,
    Variation,
    AugmentedSample,
    Results,
    Augmenter,
    ModelEvaluator,
    Category,
    Performance,
    Aggregator,
)
import countergen.types
import countergen.config



from math import exp
from typing import TYPE_CHECKING, Callable, List, Optional, Sequence, Tuple

import countergen.config
import openai
from countergen.tools.api_utils import ApiConfig
from countergen.types import Input, ModelEvaluator, Outputs, Performance

metrics = ["perplexity", "probability"]

LogProbs = float

# Return a log prob for each token of output
GenerativeModel = Callable[[Input, Outputs], Sequence[Sequence[LogProbs]]]


def api_to_generative_model(
    openai_engine: str = "ada", apiconfig: Optional[ApiConfig] = None, max_attempts: Optional[int] = 5
) -> GenerativeModel:
    """Make a GenerativeModel that uses the openai api.

    The resulting GenerativeModel takes as input an input text and possibles outputes,
    and returns the log probabilities of each tokens of each expected output.

    The GenerativeModel costs ~ len(input) * (sum of len(ouput)) tokens per call.

    If the api call fails, it will retry max_attempts times, or forever if max_attempts is None."""

    def gen_model(inp: Input, out: Outputs) -> List[List[float]]:
        apiconfig_ = apiconfig or countergen.config.apiconfig

        correct_log_probs_list = []
        for o in out:

            completion: Optional[dict] = None
            attemps = 0
            while completion is None:
                try:
                    completion = openai.Completion.create(
                        engine=openai_engine,
                        prompt=inp + o,
                        max_tokens=1,
                        stream=False,
                        echo=True,
                        logprobs=5,
                        **apiconfig_.get_config(),
                    )["choices"][0]
                except Exception as e:
                    attemps += 1
                    if max_attempts is not None and attemps >= max_attempts:
                        raise e

            token_logprobs = completion["logprobs"]["token_logprobs"]
            token_offsets = completion["logprobs"]["text_offset"]

            # Compute which tokens correspond to the output
            # If token from input & output got merged (which should happen very rarely),
            # takes into account the proability of the merged token.
            n_toks = len(token_offsets)
            start_of_output = max([i for i in range(n_toks) if token_offsets[i] <= len(inp)])

            correct_log_probs_list.append(token_logprobs[start_of_output:-1])
        return correct_log_probs_list

    return gen_model


def get_generative_model_evaluator(model: GenerativeModel, metric: str = "probability") -> ModelEvaluator:
    """Return the ModelEvaluator corresponding to the model & the metric.

    Available metrics: probability & perplexity"""

    def run(inp: Input, out: Outputs) -> Performance:
        if len(out) == 0:
            raise ValueError("Expected output should be provided for gpt models")
        if len(inp) == 0:
            raise ValueError("Empty inputs are forbidden for gpt models.")

        correct_log_probs_list = model(inp, out)

        total_prob: float = 0
        total_log_prob: float = 0
        number_of_toks: int = 0
        for correct_log_probs in correct_log_probs_list:
            total_prob += exp(sum(correct_log_probs))
            number_of_toks += len(correct_log_probs)
            total_log_prob += sum(correct_log_probs)

        if metric == "perplexity":
            return exp(-total_log_prob / number_of_toks)
        if metric == "probability":
            return total_prob
        raise ValueError(f"{metric} is not a valid metric. Choose one in {metrics}.")

    return run



import json
from ast import Or
from pathlib import Path
from typing import Any, Iterable, List, Mapping, NamedTuple, Optional, Sequence, Union, cast

import cattrs
import countergen.config
from attrs import define
from countergen.augmentation.simple_augmenter import SimpleAugmenter
from countergen.config import MODULE_PATH
from countergen.tools.utils import FromAndToJson, all_same, maybe_tqdm
from countergen.types import AugmentedSample, Augmenter, Category, Input, Outputs, Paraphraser, Sample, Variation

DEFAULT_DS_PATHS: Mapping[str, str] = {
    "doublebind-negative-1token": f"{MODULE_PATH}/data/datasets/doublebind-negative-1token.jsonl",
    "doublebind-negative": f"{MODULE_PATH}/data/datasets/doublebind-negative.jsonl",
    "doublebind-positive-1token": f"{MODULE_PATH}/data/datasets/doublebind-positive-1token.jsonl",
    "doublebind-positive": f"{MODULE_PATH}/data/datasets/doublebind-positive.jsonl",
    "doublebind-likable": f"{MODULE_PATH}/data/datasets/doublebind-likable.jsonl",
    "doublebind-unlikable": f"{MODULE_PATH}/data/datasets/doublebind-unlikable.jsonl",
    "male-stereotypes": f"{MODULE_PATH}/data/datasets/male-stereotypes.jsonl",
    "female-stereotypes": f"{MODULE_PATH}/data/datasets/female-stereotypes.jsonl",
    "tiny-test": f"{MODULE_PATH}/data/datasets/tiny-test.jsonl",
    "twitter-sentiment": f"{MODULE_PATH}/data/datasets/twitter-sentiment.jsonl",
    "hate": f"{MODULE_PATH}/data/datasets/hate-test.jsonl",
}

DEFAULT_AUGMENTED_DS_PATHS: Mapping[str, str] = {
    "doublebind-positive-paraphrased": f"{MODULE_PATH}/data/augdatasets/doublebind-positive-paraphrased.jsonl",
    "doublebind-negative-paraphrased": f"{MODULE_PATH}/data/augdatasets/doublebind-negative-paraphrased.jsonl",
    "doublebind-likable-paraphrased": f"{MODULE_PATH}/data/augdatasets/doublebind-likable-paraphrased.jsonl",
    "doublebind-unlikable-paraphrased": f"{MODULE_PATH}/data/augdatasets/doublebind-unlikable-paraphrased.jsonl",
    "male-stereotypes": f"{MODULE_PATH}/data/augdatasets/male-stereotypes.jsonl",
    "female-stereotypes": f"{MODULE_PATH}/data/augdatasets/female-stereotypes.jsonl",
    "tiny-test-aug-gender": f"{MODULE_PATH}/data/augdatasets/tiny-test.jsonl",
    "twitter-sentiment-aug-gender": f"{MODULE_PATH}/data/augdatasets/twitter-sentiment.jsonl",
    "hate-aug-muslim": f"{MODULE_PATH}/data/augdatasets/hate-test-muslims.jsonl",
    "bias-from-probs": f"{MODULE_PATH}/data/augdatasets/bias-from-probs.jsonl",
    "simple-gender": f"{MODULE_PATH}/data/augdatasets/simple-gender.jsonl",
}


@define
class SimpleAugmentedSample(Sample, AugmentedSample):
    """AugmentedSample which explicitly stores all variations."""

    variations: List[Variation] = []

    def get_variations(self) -> Sequence[Variation]:
        return self.variations

    def get_outputs(self) -> Outputs:
        return self.outputs

    @classmethod
    def from_sample(cls, s: Sample, variations: List[Variation] = []) -> "SimpleAugmentedSample":
        return SimpleAugmentedSample(s.input, s.outputs, variations)


@define
class Dataset:
    samples: List[Sample]

    @classmethod
    def from_default(cls, name: str) -> "Dataset":
        """Load one of the defaults datasets from "DEFAULT_DS_PATHS"."""
        if name not in DEFAULT_DS_PATHS:
            raise ValueError(
                f"""Default name '{name}' is not a default dataset. Choose one in {list(DEFAULT_DS_PATHS.keys())}"""
            )
        return Dataset.from_jsonl(DEFAULT_DS_PATHS[name])

    @classmethod
    def from_jsonl(cls, path: str) -> "Dataset":
        """Load a dataset from a jsonl file.

        Expected format of each line:
        {"input": "<INP>", "outputs": ["<OUT1>", "<OUT2>", ...]}
        where you have at least one accepted output per input."""
        with Path(path).open("r", encoding="utf-8") as f:
            data = [json.loads(line) for line in f]
            samples = []
            for d in data:
                samples.append(Sample.from_json_dict(d))
        return Dataset(samples)

    def augment(self, converters: Iterable[Augmenter]) -> "AugmentedDataset":
        """Use the augmenters to generate variations of this dataset."""
        return generate_all_variations(converters, self)


@define
class AugmentedDataset:
    samples: List[SimpleAugmentedSample]

    @classmethod
    def from_default(cls, name: str) -> "AugmentedDataset":
        """Load one of the defaults datasets from "DEFAULT_AUGMENTED_DS_PATHS"."""
        if name not in DEFAULT_AUGMENTED_DS_PATHS:
            raise ValueError(
                f"Default name '{name}' is not a default augmented dataset. Choose one in {list(DEFAULT_AUGMENTED_DS_PATHS.keys())}"
            )
        return AugmentedDataset.from_jsonl(DEFAULT_AUGMENTED_DS_PATHS[name])

    @classmethod
    def from_jsonl(cls, path: str) -> "AugmentedDataset":
        with Path(path).open("r", encoding="utf-8") as f:
            data = [json.loads(line) for line in f]
            samples = []
            for d in data:
                samples.append(SimpleAugmentedSample.from_json_dict(d))
        return AugmentedDataset(samples)

    def save_to_jsonl(self, path: str):
        with Path(path).open("w", encoding="utf-8") as f:
            for sample in self.samples:
                json.dump(sample.to_json_dict(), f)
                f.write("\n")

    def to_unbiased_dataset(self) -> Dataset:
        """Make each variation a new input."""
        new_samples: List[Sample] = []

        for aug_sample in self.samples:
            for variation in aug_sample.get_variations():
                new_samples.append(Sample(input=variation.text, outputs=aug_sample.outputs))

        return Dataset(new_samples)


def generate_variations(variation: Variation, augmenter: Augmenter) -> List[Variation]:
    if isinstance(augmenter, Paraphraser):
        return generate_paraphrase(variation, cast(Paraphraser, augmenter))

    new_variations = [
        Variation(augmenter.transform(variation.text, category), variation.categories + (category,))
        for category in augmenter.categories
    ]
    if not all_same([v.text for v in new_variations]):
        return new_variations
    else:
        return [variation]


def generate_paraphrase(variation: Variation, augmenter: Paraphraser) -> List[Variation]:
    new_text = augmenter.transform(variation.text)
    if new_text == variation.text:
        return [variation]
    else:
        return [variation, Variation(new_text, variation.categories)]


def generate_all_variations(augmenters: Iterable[Augmenter], ds: Dataset) -> AugmentedDataset:
    """Apply each augmenter to each sample of the dataset for each available target category.

    It first replaces samples with the transformed samples through the first augmenter,
    then replaces these with samples transformed through the second augmenter, and so on.
    Return an number of variations that can be exponential in the number of augmenters.

    Remove duplicates.

    If an augmenter is a paraphrase, keep the original input too."""

    augmented_samples = []

    for sample in maybe_tqdm(ds.samples, countergen.config.verbose >= 2):
        variations = [Variation(sample.input, ())]
        for augmenter in augmenters:
            new_variations = []
            for v in variations:
                new_variations += generate_variations(v, augmenter)
            variations = new_variations
        augmented_samples.append(SimpleAugmentedSample.from_sample(sample, variations))
    return AugmentedDataset(augmented_samples)



from typing import List, Optional, Union

import fire  # type: ignore

from countergen import (
    DEFAULT_CONVERTERS_PATHS,
    AugmentedDataset,
    Dataset,
    SimpleAugmenter,
    api_to_generative_model,
    evaluate_and_print,
    get_generative_model_evaluator,
)
from countergen.augmentation.paraphraser import LlmParaphraser
from countergen.types import Augmenter


def _augment(load_path: str, save_path: str, *augmenters_desc: str):
    """Add counterfactuals to the dataset and save it elsewhere.

    Args
    - load-path: the path of the dataset to augment
    - save-path: the path where the augmenter dataset will be save
    - augmenters: a list of ways of converting a string to another string.
                  * If it ends with a .json, assumes it's a the path to a file containing
                  instructions to build a converter. See the docs [LINK] for more info.
                  * Otherwise, assume it is one of the default augmenters: either 'gender' or 'west_v_asia' or 'paraphrase'
                  * If no converter is provided, default to 'gender'

    Example use:
    - countergen augment LOAD_PATH SAVE_PATH gender west_v_asia
    - countergen augment LOAD_PATH SAVE_PATH CONVERTER_PATH
    - countergen augment LOAD_PATH SAVE_PATH gender CONVERTER_PATH
    - countergen augment LOAD_PATH SAVE_PATH
    """

    if not augmenters_desc:
        augmenters_desc = ("gender",)

    augmenters: List[Augmenter] = []
    for c_str in augmenters_desc:
        if c_str.endswith(".json"):
            augmenter: Augmenter = SimpleAugmenter.from_json(c_str)
        elif c_str in DEFAULT_CONVERTERS_PATHS:
            augmenter = SimpleAugmenter.from_default(c_str)
        elif c_str == "paraphrase":
            augmenter = LlmParaphraser()
        else:
            print(f"{c_str} is not a valid augmenter name.")
            return
        augmenters.append(augmenter)
    ds = Dataset.from_jsonl(load_path)
    aug_ds = ds.augment(augmenters)
    aug_ds.save_to_jsonl(save_path)
    print("Done!")


def _evaluate(
    load_path: str,
    save_path: Optional[str] = None,
    model_name: Optional[str] = None,
):
    """Evaluate the provided model.

    Args
    - load-path: the path to the augmented dataset
    - save-path: Optional flag. If present, save the results to the provided path. Otherwise, print the results
    - model-name: Optional flag. Use the model from the openai api given after the flag, or ada is none is provided

    Note: the augmented dataset should match the kind of network you evaluate! See the docs [LINK] for more info.

    Example use:
    - countergen evaluate LOAD_PATH SAVE_PATH
      (use ada and save the results)
    - countergen evaluate LOAD_PATH  --model-name davinci
      (use GPT-3 and print the results)
    """

    ds = AugmentedDataset.from_jsonl(load_path)
    model_api = api_to_generative_model() if model_name is None else api_to_generative_model(model_name)
    model_ev = get_generative_model_evaluator(model_api)

    evaluate_and_print(ds.samples, model_ev)

    if save_path is not None:
        print("Done!")


def _overwrite_fire_help_text():  # type: ignore
    import inspect
    import fire  # type: ignore
    from fire import inspectutils  # type: ignore
    from fire.core import _ParseKeywordArgs  # type: ignore

    # Replace the default help text by the __doc__
    def NewHelpText(component, trace=None, verbose=False):
        if callable(component):
            return component.__doc__
        elif isinstance(component, dict):
            docs = {k: v.__doc__.split("\n")[0] for k, v in component.items()}
            return "COMMANDS\n" + "\n".join(f"  {k}\n   {doc}" for k, doc in docs.items())
        else:
            return ""

    fire.helptext.HelpText = NewHelpText

    # Remove the INFO line
    def _NewIsHelpShortcut(component_trace, remaining_args):
        show_help = False
        if remaining_args:
            target = remaining_args[0]
            if target in ("-h", "--help"):
                # Check if --help would be consumed as a keyword argument, or is a member.
                component = component_trace.GetResult()
                if inspect.isclass(component) or inspect.isroutine(component):
                    fn_spec = inspectutils.GetFullArgSpec(component)
                    _, remaining_kwargs, _ = _ParseKeywordArgs(remaining_args, fn_spec)
                    show_help = target in remaining_kwargs
                else:
                    members = dict(inspect.getmembers(component))
                    show_help = target not in members
        if show_help:
            component_trace.show_help = True
            # [Where the INFO line was printed]
        return show_help

    fire.core._IsHelpShortcut = _NewIsHelpShortcut


def run():
    _overwrite_fire_help_text()
    fire.Fire(
        {
            "augment": _augment,
            "evaluate": _evaluate,
        },
    )



from math import exp, log2
from typing import Any, Callable, Dict, Iterable, Mapping, Optional, Sequence, TypeVar, Tuple


def mean(l: Sequence[float]) -> float:
    return sum(l) / len(l)


def geometric_mean(l: Sequence[float]) -> float:
    return 2 ** (mean(list(map(log2, l))))


def perplexity(log_probs: Sequence[float]):
    """Take in natural log probs, returns (average) perplexity"""
    return exp(mean(log_probs))



from typing import Mapping, Optional
import matplotlib.pyplot as plt  # type: ignore
import numpy as np

TypeName = str
CategoryName = str


def plot_mutli_bars(
    scores_by_type_by_cat: Mapping[TypeName, Mapping[CategoryName, float]],
    margin: float = 0.25,
    xlabel: str = "",
    ylabel: str = "",
    title: str = "",
    err_by_type_by_cat: Optional[Mapping[TypeName, Mapping[CategoryName, float]]] = None,
):
    type_names = list(scores_by_type_by_cat.keys())
    n_types = len(type_names)
    assert n_types > 0

    categories = list(scores_by_type_by_cat[type_names[0]].keys())
    n_categories = len(categories)
    assert n_categories > 0

    ind = np.arange(n_types)
    width = (1 - margin) / n_categories

    for i, category in enumerate(categories):
        bar_pos = ind + width * i - width * (n_categories - 1) / 2
        perfs = [scores_by_type_by_cat[type_name][category] for type_name in type_names]
        plt.bar(bar_pos, perfs, width, label=category)
        if err_by_type_by_cat is not None:
            plt.errorbar(
                bar_pos, perfs, yerr=[err_by_type_by_cat[type_name][category] for type_name in type_names], color="k"
            )

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)

    plt.xticks(ind, type_names)
    plt.legend(loc="best")
    plt.show()



import abc
from typing import (
    Callable,
    Generic,
    Iterable,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    TextIO,
    TypeVar,
    List,
    Tuple,
    Union,
)

from attrs import define
from countergen.tools.utils import FromAndToJson

Input = str  # The input to an NLP mode
Outputs = List[str]  # The different acceptable outputs of the NLP, string label or number, but in string format


@define
class Sample(FromAndToJson):
    """An input associated with its possible outputs."""

    input: Input
    outputs: Outputs = []


Category = str  # The different kinds of data produced by augmenters.


class Augmenter(metaclass=abc.ABCMeta):
    """Objects that can transform an input to another input from a target category.

    Also include objects which can do paraphrase."""

    @abc.abstractproperty
    def categories(self) -> Tuple[Category, ...]:
        """Categories the augmenter can output."""
        ...

    @abc.abstractmethod
    def transform(self, inp: Input, to: Category) -> Input:
        """Transform the input to make it a member of the target category.

        Return the input unchanged if the input is already a member of the target category or if
        the transformation is not applicable."""
        ...


class Paraphraser(Augmenter):
    """Objects that can transform an input to another input with the same meaning."""

    @abc.abstractmethod
    def transform(self, inp: Input, to: Category = "") -> Input:
        """Transform the input to make it another input with the same meaning.

        Ignore the target category."""
        ...

    @property
    def categories(self) -> Tuple[Category, ...]:
        return ()


@define
class Variation(FromAndToJson):
    """An input associated with the different categories it belongs to."""

    text: Input
    categories: Tuple[Category, ...]


class AugmentedSample(metaclass=abc.ABCMeta):
    """Hold different variations of a sample with the same meaning, which should have the same expected outputs."""

    @abc.abstractproperty
    def input(self) -> Input:
        """The original input."""
        ...

    @abc.abstractproperty
    def outputs(self) -> Outputs:
        """The expected outputs."""
        ...

    @abc.abstractmethod
    def get_variations(self) -> Sequence[Variation]:
        """The variations of the original input, which still should procude one of the expected outputs."""
        ...


# The performance over outputs, or the performance on every output
# usually between zero & one (one is better)
Performance = Union[float, List[float]]

VariationResult = Tuple[Performance, Tuple[Category, ...]]
SampleResults = Sequence[VariationResult]
Results = Sequence[SampleResults]

# Callable that returns the performance of a model given an input and expected outputs.
ModelEvaluator = Callable[[Input, Outputs], Performance]

Aggregate = TypeVar("Aggregate")


class Aggregator(Generic[Aggregate], metaclass=abc.ABCMeta):
    """Objects that can aggregate, and optionally save, load and display the performances of models.

    The aggregate can be of any type."""

    @abc.abstractmethod
    def __call__(self, performances: Results) -> Aggregate:
        """Return an aggregate of the performances.

        The performances is a list of the results for each sample:
        for each sample, the performance and the categories of each variations
        are given."""
        ...

    def save_aggregation(self, aggregate: Aggregate, file: Optional[TextIO] = None):
        """Save the aggregate to a file."""
        raise NotImplementedError(f"{self.__class__.__name__} can't save aggregates.")

    def load_aggregation(self, file: TextIO) -> Aggregate:
        """Load the aggregate from a file."""
        raise NotImplementedError(f"{self.__class__.__name__} can't load aggregates.")

    def display(self, aggregates: Mapping[str, Aggregate]):
        """Display and compare the aggregates of different models/datasets.

        The keys of the aggregates map should be a short description of what differs between
        the aggregates (model names if models are different, dataset names if the data changes, ...)"""
        print(aggregates)



from typing import Dict, Optional
from attrs import define
from countergen.tools.utils import FromAndToJson


@define
class ApiConfig(FromAndToJson):
    """Hold API key and API URL"""

    key: Optional[str] = None
    base_url: Optional[str] = None

    def get_config(self) -> Dict[str, str]:
        """Return the argument the openai module needs."""

        if self.key is None:
            raise RuntimeError(
                "Please provide openai key to use its api! Use `countergen.config.apiconfig.key = YOUR_KEY`"
            )
        if self.base_url is None:
            raise RuntimeError(
                "Please provide openai base url to use its api! Use `countergen.config.apiconfig.base_url = BASE_URL`"
            )

        return {
            "api_key": self.key,
            "api_base": self.base_url,
        }



from typing import Dict, Optional, Tuple
from countergen.tools.api_utils import ApiConfig

import openai
from attrs import define
from countergen.tools.utils import estimate_paraphrase_length
from countergen.types import Augmenter, Category, Input, Paraphraser
import countergen.config

# Examples from https://www.pragnakalp.com/intent-classification-paraphrasing-examples-using-gpt-3/
DEFAULT_TEMPLATE = """Paraphrase the original passage.

Original: {Searching a specific search tree for a binary key can be programmed recursively or iteratively.}
Paraphrase: {Searching a specific search tree according to a binary key can be recursively or iteratively programmed.}

Original: {It was first released as a knapweed biocontrol in the 1980s in Oregon , and it is currently established in the Pacific Northwest.}
Paraphrase: {It was first released as Knopweed Biocontrol in Oregon in the 1980s , and is currently established in the Pacific Northwest.}

Original: {4-OHT binds to ER , the ER / tamoxifen complex recruits other proteins known as co-repressors and then binds to DNA to modulate gene expression.}
Paraphrase: {The ER / Tamoxifen complex binds other proteins known as co-repressors and then binds to DNA to modulate gene expression.}

Original: {In mathematical astronomy, his fame is due to the introduction of the astronomical globe, and his early contributions to understanding the movement of the planets.}
Paraphrase: {His fame is due in mathematical astronomy to the introduction of the astronomical globe and to his early contributions to the understanding of the movement of the planets.}

Original: {__input__}
Paraphrase: {"""


@define
class LlmParaphraser(Paraphraser):
    prompt_template: str = DEFAULT_TEMPLATE
    engine: str = "text-davinci-003"
    apiconfig: Optional[ApiConfig] = None

    def transform(self, inp: Input, to: Category = "") -> Input:
        apiconfig = self.apiconfig or countergen.config.apiconfig

        prompt = self.prompt_template.replace("__input__", inp)

        completion = openai.Completion.create(
            engine=self.engine,
            prompt=prompt,
            max_tokens=estimate_paraphrase_length(inp),
            temperature=1,
            top_p=0.7,  # LLM-D has top_k=40, but not available
            stream=False,
            **apiconfig.get_config(),
        )["choices"][0]["text"]

        return completion.split("}")[0]



from pathlib import Path
from typing import Iterable, Mapping, TypeVar
from countergen.evaluation.aggregators import PerformanceStatsPerCategory
import countergen.config

from countergen.types import (
    AugmentedSample,
    Category,
    ModelEvaluator,
    Performance,
    Results,
    Aggregator,
)
from countergen.tools.utils import maybe_tqdm
from countergen.tools.math_utils import mean


T = TypeVar("T")


def compute_performances(samples: Iterable[AugmentedSample], model: ModelEvaluator) -> Results:
    """Compute and return the performance of each variation in each sample.

    Preserves the order in which samples and variations are given."""
    performances = []
    for sample in maybe_tqdm(samples, countergen.config.verbose >= 2):
        performance = [
            (model(variation.text, sample.outputs), variation.categories) for variation in sample.get_variations()
        ]
        performances.append(performance)
    return performances


def evaluate(
    samples: Iterable[AugmentedSample],
    model: ModelEvaluator,
    aggregator: Aggregator[T] = PerformanceStatsPerCategory(),
) -> T:
    """Return the aggregate of the performance of each variation in each sample."""
    return aggregator(compute_performances(samples, model))


def evaluate_and_print(
    samples: Iterable[AugmentedSample],
    model: ModelEvaluator,
    aggregator: Aggregator[T] = PerformanceStatsPerCategory(),
):
    aggregator.save_aggregation(evaluate(samples, model, aggregator))


def evaluate_and_save(
    samples: Iterable[AugmentedSample],
    model: ModelEvaluator,
    path: str,
    aggregator: Aggregator[T] = PerformanceStatsPerCategory(),
    also_print: bool = True,
):
    with Path(path).open("w", encoding="utf-8") as f:
        perfs = evaluate(samples, model, aggregator)
        aggregator.save_aggregation(perfs, file=f)
        if also_print:
            aggregator.save_aggregation(perfs)



import json
from collections import defaultdict
from pathlib import Path
from random import choice
from typing import Any, Callable, DefaultDict, Dict, Iterable, List, Mapping, Sequence, Tuple

import re
from attrs import define

from countergen.types import Category, Augmenter, Input
from countergen.tools.utils import other
from countergen.config import MODULE_PATH

DEFAULT_CONVERTERS_PATHS: Mapping[str, str] = {
    "gender": f"{MODULE_PATH}/data/converters/gender.json",
    "west_v_asia": f"{MODULE_PATH}/data/converters/west_v_asia.json",
}


@define
class ConversionDataset:
    categories: Tuple[Category, Category]
    correspondances: List[Tuple[List[str], List[str]]]

    @classmethod
    def from_json(cls, json_dict: Mapping[str, Any]) -> "ConversionDataset":
        categories = tuple(json_dict["categories"])
        correspondances_maps = json_dict["correspondances"]
        cat_0, cat_1 = categories
        correspondances = []
        for m in correspondances_maps:
            correspondances.append((m[cat_0], m[cat_1]))
        return ConversionDataset(categories, correspondances)  # type: ignore


Transformation = Callable[[str], str]
CorrespondanceDict = Dict[Category, DefaultDict[str, List[str]]]
DEFAULT_TRANSFORMATIONS = [
    lambda s: s.lower(),
    lambda s: s.upper(),
    lambda s: s.capitalize(),
]


@define
class SimpleAugmenter(Augmenter):
    """Augmenter that does word substituion between its two categories.

    "words" are defined by the word_regex expression."""

    categories: Tuple[Category, Category]
    correspondance_dict: CorrespondanceDict
    word_regex: str = r"([A-Za-zÀ-ÖØ-öø-ÿ]+\-[A-Za-zÀ-ÖØ-öø-ÿ]+)|[A-Za-zÀ-ÖØ-öø-ÿ]+"

    @classmethod
    def from_default(cls, name: str = "gender") -> "SimpleAugmenter":
        """Load one of the defaults datasets from "DEFAULT_CONVERTERS_PATHS"."""
        if name not in DEFAULT_CONVERTERS_PATHS:
            raise ValueError(
                f"""Default name '{name}' is not a default simple augmenter. Choose one in {list(DEFAULT_CONVERTERS_PATHS.keys())}"""
            )
        return SimpleAugmenter.from_json(DEFAULT_CONVERTERS_PATHS[name])

    @classmethod
    def from_json(
        cls,
        path: str,
        transformations: Iterable[Transformation] = DEFAULT_TRANSFORMATIONS,
    ) -> "SimpleAugmenter":
        """Create the SimpleAugmenter corresponding to the substitution rules defined in a json file.
        The json file should be of the following form:
        {
        "categories": ["<CATA>","<CATB>"],
        "correspondances": [
        {"<CATA>": ["<WA1>", "<WA2>", ...],"<CATB>": ["<WB1>", "<WB2>", ...]},
        ...]
        }
        where <CATA> and <CATB> are the two categories, and <WA...> can be replaced by <WB...>
        to go from category <CATA> to category <CATB>."""

        with Path(path).open("r", encoding="utf-8") as f:
            json_dict: Mapping[str, Any] = json.loads(f.read())
            return SimpleAugmenter.from_ds(ConversionDataset.from_json(json_dict), transformations)

    @classmethod
    def from_ds(
        cls,
        ds: ConversionDataset,
        transformations: Iterable[Transformation] = DEFAULT_TRANSFORMATIONS,
    ) -> "SimpleAugmenter":
        correspondance_dict: CorrespondanceDict = {}
        for c in ds.categories:
            correspondance_dict[c] = defaultdict(lambda: [])

        for correspondance in ds.correspondances:
            correspondance_t = {
                c: {t.__code__: list(map(t, l)) for t in transformations} for c, l in zip(ds.categories, correspondance)
            }

            for c, l in zip(ds.categories, correspondance):
                for word in l:
                    for t in transformations:
                        correspondance_dict[c][t(word)] += correspondance_t[other(ds.categories, c)][t.__code__]

        return SimpleAugmenter(ds.categories, correspondance_dict)

    def transform(self, inp: Input, to: Category) -> Input:
        from_category = other(self.categories, to)
        p = re.compile(self.word_regex)
        new_inp_pieces = []
        end = 0
        for group in p.finditer(inp):
            word = inp[group.start() : group.end()]
            if self.correspondance_dict[from_category][word]:
                new_inp_pieces.append(inp[end : group.start()])
                new_inp_pieces.append(choice(self.correspondance_dict[from_category][word]))
                end = group.end()

        new_inp_pieces.append(inp[end:])
        return "".join(new_inp_pieces)



import pathlib
from setuptools import setup, find_packages  # type: ignore

"""
To upload the package, run
python setup.py sdist
twine upload --repository-url https://upload.pypi.org/legacy/ dist/countergen-VERSION.tar.gz
"""

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()
VERSION = (HERE / "VERSION").read_text()
REQUIREMENTS = (HERE / "requirements.txt").read_text().splitlines()

setup(
    name="countergen",
    version=VERSION,
    description="A counterfactual dataset generator to evaluate language models.",
    long_description=README,
    long_description_content_type="text/markdown",
    author="SaferAI",
    author_email="saferai.audit@gmail.com",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    url="https://github.com/FabienRoger/Countergen",
    python_requires=">=3.7",
    include_package_data=True,
    install_requires=REQUIREMENTS,
    entry_points={
        "console_scripts": [
            "countergen = countergen:run",
        ],
    },
)
